import * as THREE from 'https://cdn.skypack.dev/three@0.128.0'
import { OrbitControls } from 'https://cdn.skypack.dev/three@0.128.0/examples/jsm/controls/OrbitControls.js'

let socket = io()
let serial_data

//10-bit value that can come in from the phototransistor
const max_value = 1023
const num_bars = max_value+1 //Math.round(max_value/10)

//Color scale
const colors = ['#423b8a', '#6a469b', '#9d4ba4', '#d744a6', '#ff587a', '#fc8c36', '#efb605'].reverse()
const scale_color = d3.scaleLinear()
    .domain(d3.range(colors.length).map(d => { return d/(colors.length-1) * max_value}))
    .range(colors)
    .clamp(true)

//Create positions along Hilbert curve
const hilbert_order = 5 //32 * 32 square -> 1024 rectangles
const size = 2 ** hilbert_order //the size of one side
const hilbert = []
for(let i = 0; i < num_bars; i++){
    hilbert.push(hilbertCurve.indexToPoint(i, hilbert_order))
}//for i
const scale = 0.5

//Set the correct size of the "chart" div
const container = document.getElementById("chart")
const width = 800
const height = 800
container.style.width = `${width}px`
container.style.height = `${height}px`

///////////////// Setup the Three.js scene ////////////////

//Create the Three.js scene
const scene = new THREE.Scene()

//Setup the canvas and webGL renderer and add it to the HTML
const renderer = new THREE.WebGLRenderer({ antialias: true})
renderer.setSize( width, height )
renderer.setPixelRatio( window.devicePixelRatio )
renderer.setClearColor("#fff", 1) //background color
renderer.shadowMap.enabled = true //otherwise none of the shadow settings will work
container.appendChild( renderer.domElement )

///////////////////// Setup the camera ////////////////////

//Set up a camera for the viewer to see into the scene
//Field-of-view, aspect ration, near & far clipping place
const camera = new THREE.PerspectiveCamera( 75, 1, 0.1, 1000 )
camera.position.set(0, 0, 25)
camera.lookAt(new THREE.Vector3())

//Make it possible to move the camera's position by click-dragging and zooming
const controls = new OrbitControls(camera, renderer.domElement)

///////////////////// Add Ground Plane ////////////////////
const geo_plane = new THREE.PlaneGeometry( 100, 100, 1, 1 )
const mat_plane = new THREE.MeshStandardMaterial( { color: "#f9d4e9" } )
const floor = new THREE.Mesh( geo_plane, mat_plane )
floor.material.side = THREE.DoubleSide
floor.position.z = -0.5
floor.receiveShadow = true
scene.add( floor )

//////////////////////// Add cubes ////////////////////////
const geo_cube = new THREE.BoxGeometry()

//Create each cube and return a THREE.Mesh
const meshes = Array.from(new Array(num_bars)).map((d,i) => {
    const material = new THREE.MeshStandardMaterial({ color: scale_color(i) })

    //Create the mesh
    const cube = new THREE.Mesh(geo_cube, material)
    cube.castShadow = true
    cube.receiveShadow = true

    //Place the cube
    cube.position.x = -size/2 + hilbert[i].x
    cube.position.y = -size/2 + hilbert[i].y
    cube.position.z = -scale/2

    //Scale the cube
    cube.scale.set(scale, scale, scale)

    return cube
})//meshes

//Add meshes to the scene
meshes.forEach(m => scene.add(m))

///////////////////////// Add Text ////////////////////////

let font
let font_loaded = false
let text
let geo_text

//Standard material for each letter
const mat_font = new THREE.MeshStandardMaterial({ color: "#ffffff" })

//Load the font's json
const loader = new THREE.FontLoader()
loader.load( 'lib/source-code-pro/Source_Code_Pro_Black.json', function ( response ) {
    font = response
    font_loaded = true
})//loader

function createText(text) {
    geo_text = new THREE.TextGeometry( text, {
        font: font,
        size: 7,
        height: 1,
        curveSegments: 12,

        bevelEnabled: true,
        bevelThickness: 0.25,
        bevelSize: 0.1,
        bevelOffset: 0,
        bevelSegments: 5
    })

    geo_text.computeBoundingBox()
    const offset_x = ( geo_text.boundingBox.max.x - geo_text.boundingBox.min.x )/2
    const offset_y = ( geo_text.boundingBox.max.y - geo_text.boundingBox.min.y )/2

    //Create and position mesh
    const mesh = new THREE.Mesh( geo_text, mat_font )
    mesh.position.x = -offset_x
    mesh.position.y = -offset_y
    mesh.position.z = 1

    mesh.castShadow = true
    mesh.receiveShadow = true

    return mesh
}//function createText

////////////////////// Create Lights //////////////////////

//Ambient light
const light = new THREE.AmbientLight( "#fbe7f2", 1 )
scene.add( light )

//Directional light
const light_direction = new THREE.DirectionalLight("#ffffff", 0.8)
light_direction.castShadow = true
light_direction.position.set(-20, -15, 20)
light_direction.target.position.set(0, 0, 0)

//Increase frustum of light and shadow casting
const d = 20
light_direction.shadow.camera.left = -d
light_direction.shadow.camera.right = d
light_direction.shadow.camera.top = d
light_direction.shadow.camera.bottom = -d
//Increase resolution of shadow
light_direction.shadow.camera.near = 1
light_direction.shadow.camera.far = 50
light_direction.shadow.mapSize.x = 2048
light_direction.shadow.mapSize.y = 2048

scene.add(light_direction)
scene.add(light_direction.target)

// //Show helper to show direction and location of light
// const helper = new THREE.DirectionalLightHelper(light_direction)
// scene.add(helper)

///////////////////////// Animate /////////////////////////

function animate() {
    if(serial_data) {
        //Possibly change the z-position of each cube
        meshes.forEach((mesh,i) => {
            let z
            if(i < serial_data) z = -scale/2
            else z = -scale*1.6

            //Use LERP to (possibly) move the position of the cubes above or below the plane
            if(z !== mesh.position.z) mesh.position.lerp(new THREE.Vector3(-size/2 + hilbert[i].x, -size/2 + hilbert[i].y, z), 0.1)
        })//forEach
    
        //Change the color of the floor
        floor.material.color.set(scale_color(serial_data))

        //Update the text
        if(font_loaded) {
            //Clean up, otherwise the meshes will keep gobbling up the memory
            //This still doesn't fix it all, but the memory increases more slowly
            if(text) {
                scene.remove(text)
                geo_text.dispose()
                text.material.dispose()
                text.geometry.dispose()
                renderer.renderLists.dispose()
            }//if
            //Add text
            text = createText(serial_data)
            scene.add(text)
        }//if
    }//if

    //Update scene
    controls.update()
	renderer.render( scene, camera )

    requestAnimationFrame( animate )
}//function animate

animate()

///////////////////////// Get Data ////////////////////////
//Get the data from the serial port and update the visual
socket.on('connect', function () {
    socket.on('parsed-data', function (data) {
        //console.log(data)
        serial_data = data
    })//socket.on
})//socket.on
